#!/usr/bin/env python3

"""

    NaviX is a server with the target "bring the job to the data".
    Therefore NaviX is called from HTCondor via translate- and update-hook, this hooks send the line "inputfiles" to
    NaivX and in case of the update-hook also a timestamp. NaviX requests the files on the CMSD-Server and responds
    to HTCondor with a Rank and also modifies the requirement of the job.

"""
from os import environ
from multiprocessing import active_children

from navix.to_hook_networking import *
from navix.compile import *
from navix.ClientProcess import ClientProcess
from navix.BufferingService import BufferingService
from sys import argv
from navix.logging import *


if __name__ == '__main__':

    #####################################################################
    #
    # Set simple logger for startup only -> should never log anything
    # if everything works on startup
    #
    #####################################################################
    # log file path
    logFileS = 'NaviX_start_error.log'

    # create logger
    start_log = logging.getLogger(logFileS)
    start_log.setLevel(logging.INFO)

    # create console handler and set level to debug
    console_log = logging.StreamHandler()
    console_log.setLevel(logging.DEBUG)
    # create formatter
    formatter_log = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s', datefmt='%Y/%m/%d %H:%M:%S')
    # add formatter to ch
    console_log.setFormatter(formatter_log)
    # add ch to logger
    start_log.addHandler(console_log)
    log_handler = RotatingFileHandler(logFileS, mode='a', maxBytes=1 * 1024 * 1024,
                                      backupCount=3, encoding=None, delay=0)
    log_handler.setFormatter(formatter_log)
    start_log.addHandler(log_handler)

    #####################################################################
    #
    # Read config and set logger level
    #
    #####################################################################
    try:
        logging_level_param = argv[1]
    except Exception as err:
        # No logging level is set
        logging_level_param = False

    try:
        settings = Settings.get_instance(start_log, logging_level_param)
    except Exception as err:
        start_log.error("ERROR: Failed to initialize config object -> Exit"
                        + "Python-ERROR: " + str(err))
        exit()

    #####################################################################
    #
    # Initialize logging class
    #
    #####################################################################
    try:
        logging = Logging()
    except Exception as err:
        start_log.error("ERROR: Failed to initialize logging object -> Exit"
                        + "Python-ERROR: " + str(err))

    # inform about logging level:
    # noinspection PyUnboundLocalVariable
    logging.queue_logging.put("_INFO: Log level is " + settings.log_level)

    logging.queue_logging.put("_INFO: Server Start\n"
                              + "######################################################################\n"
                              + "#                                                                    #\n"
                              + "#                     NaviX STARTING                                 #\n"
                              + "#                                                                    #\n"
                              + "######################################################################\n")

    #####################################################################
    #
    # Compile and export hook
    #
    #####################################################################

    if settings.compile_hook:
        try:
            start_compile_hook(logging.queue_logging, settings)
        except Exception as err:
            logging.queue_logging.put("ERROR: Failed to compile hooks -> Exit"
                                      + "Python-ERROR: " + str(err))
            sleep(0.5)    # sleep -> waiting for logger
            exit()

    #####################################################################
    #
    # Set variable of xrootd for NaviX
    #
    #####################################################################
    # TODO has no effect at the moment - fix this
    try:
        environ["XRD_TIMEOUTRESOLUTION"] = str(settings.pyxrootd_timeout)
        logging.queue_logging.put("DEBUG: Successfully set timeout resolution for XrootD to "
                                  + str(settings.pyxrootd_timeout) + "s")
    except Exception as err:
        logging.queue_logging.put("ERROR: Failed to set timeout resolution for XrootD! -> stopping server\n"
                                  + "Python ERROR: " + str(err))
        sleep(0.5)  # sleep -> waiting for logger
        exit()

    #####################################################################
    #
    # Open Server socket
    #
    #####################################################################
    logging.queue_logging.put("DEBUG: Trying to open server socket")
    try:
        sock = open_socket(settings.port, logging.queue_logging, "INIT", settings.host)
        logging.queue_logging.put("DEBUG: Successfully opened server socket")
    except Exception as err:
        logging.queue_logging.put("ERROR: Failed to open server socket! -> stopping server\nPort available?\n"
                                  + "Python ERROR: " + str(err))
        sleep(0.5)  # sleep -> waiting for logger
        exit()

    #####################################################################
    #
    # Starting buffering service in separated process
    #
    #####################################################################
    logging.queue_logging.put("DEBUG: Trying to spawn buffering service")
    try:
        # start new process 1st argument -> function to run,
        # second are arguments to the function.
        buffering_service_process = Process(target=BufferingService(logging, settings).run, args=())
        # buffering_service_process.daemon = True
        buffering_service_process.start()
        logging.queue_logging.put("DEBUG: Successfully spawned buffering service")
    except Exception as err:
        logging.queue_logging.put("ERROR: Failed to spawn buffering service! -> stopping server\n"
                                  + "Python ERROR: " + str(err))
        sleep(0.5)  # sleep -> waiting for logger
        exit()

    #####################################################################
    #
    # Server waiting for incoming requests -> starting thread per client
    #
    #####################################################################
    # now keep talking with the client
    logging.queue_logging.put("_INFO: Server is started and know listening for hook connections.")
    thread_number = 0

    # needed to clean up subprocess (see except socket.timeout )
    # noinspection PyUnboundLocalVariable
    sock.settimeout(5)
    try:
        while True:
            try:
                # wait to accept a connection - blocking call
                conn, address = sock.accept()
                logging.queue_logging.put("DEBUG: Connection attempt from " + address[0] + ":" + str(address[1]))
                # start new process 1st argument -> function to run,
                # second are arguments to the function.
                process = Process(target=ClientProcess, args=(conn, logging, thread_number, settings))
                process.daemon = True
                process.start()
                thread_number += 1

                result = active_children()
                logging.queue_logging.put("DEBUG: Current running processes: " + str(len(result)))
            except socket.timeout:
                # clean remaining zombies (sub processes want to return value, but the mother process is blocked by
                # sock.accept)
                result = active_children()
                logging.queue_logging.put("DEBUG: Current running processes: " + str(len(result)))
    except (KeyboardInterrupt, SystemExit):
        sock.close()
        logging.queue_logging.put("_INFO: Shutdown of server!")
        sleep(1)