import threading
import multiprocessing as mp
from time import sleep
from time import time


########################################################################################################################
#
# The buffering service is a daemon that periodically collects information and dumps it into the database
#
########################################################################################################################


def get_htcondor_information(process_id, output, logging):
    """
    Collect information from HTCondor infrastructure
    :param output: dictionary containing collected HTCondor information that is passed to the output queue
    :param process_id: ID of the sub-process to sort output
    :param logging: default logging queue
    :return:
    """
    try:
        # TODO: Write code here
        # self.queue_logging.put("DEBUG: Trying to collect HTCondor information for buffering service")
        sleep(3)
        htcondor_dict = {'htcondor': 'not connected yet'}
        # self.queue_logging.put("DEBUG: Successfully collected HTCondor information for buffering service")
        output.put((process_id, htcondor_dict))
    except Exception as err:
        logging.put("ERROR: Failed to get HTCondor information\n" + "Python ERROR: " + str(err))
        raise


def get_xrootd_information(process_id, output, logging):
    """
    Collect information from XRootD infrastructure
    :param output: dictionary containing collected XRootD information that is passed to the output queue
    :param process_id: ID of the sub-process to sort output
    :param logging: default logging queue
    :return:
    """
    try:
        # TODO: Write code here
        # self.queue_logging.put("DEBUG: Trying to collect XRootD information for buffering service")
        sleep(2)
        xrootd_dict = {'xrootd': 'not connected yet'}
        # self.queue_logging.put("DEBUG: Successfully collected XRootD information for buffering service")
        output.put((process_id, xrootd_dict))
    except Exception as err:
        logging.put("ERROR: Failed to get XRootD information\n" + "Python ERROR: " + str(err))
        raise


class BufferingService:
    """
    Collects information from HTCondor, XRootD and NaviX periodically and dumps it into the database
    """
    def __init__(self, logging, settings):
        """
        Initialize buffering service and start repeated updating of database information
        :param logging: queue for logging (multiprocessing compatible) -> logging in one file NaviX.log
        :param settings: object containing all important settings
        :return:
        """
        # list of all attributes
        self.queue_logging = logging.queue_logging
        self.settings = settings
        self.__htcondor_dict = dict()
        self.__xrootd_dict = dict()

        # run process
        # self.run()

    def run(self):
        """
        Spawn processes for getting information and writing into database
        :return:
        """
        try:
            # create queue for getting results from sub-processes
            output = mp.Queue()
            ticker = threading.Event()
            while not ticker.wait(int(self.settings.information_update_interval)):
                start = time()
                self.queue_logging.put("DEBUG: Trying to collect information for buffering service")
                # create list of all sub-processes
                process_list = [mp.Process(target=get_htcondor_information, args=(1, output, self.queue_logging)),
                                mp.Process(target=get_xrootd_information, args=(2, output, self.queue_logging))]
                # Start the sub-processes (i.e. calculate the random number lists)
                [process.start() for process in process_list]
                # Ensure all of the sub-processes have finished
                [process.join() for process in process_list]
                # Get results back from sub-processes
                output_list = [output.get() for process in process_list]
                output_list.sort()
                [self.__htcondor_dict, self.__xrootd_dict] = [output_value[1] for output_value in output_list]
                if not all([process.exitcode == 0 for process in process_list]):
                    self.queue_logging.put("INFO: Collecting information for buffering service crashed, "
                                           "skipping database update")
                else:
                    self.queue_logging.put("DEBUG: Successfully collected information for buffering service")
                    # Write to information into database
                    self.write_to_database()
                    # sleep(10)
                self.queue_logging.put("DEBUG: Collecting-Process took: " + str(round(time() - start, 1)) + " seconds")
        except Exception as err:
            self.queue_logging.put("ERROR: Uncaught exception in buffering service -> Kill process\n"
                                   + "Python-ERROR: " + str(err))
            sleep(0.1)  # sleep to flush to queue
            raise

    def write_to_database(self):
        """
        Write collected information into database
        :return:
        """
        # TODO: Write code here
        try:
            self.queue_logging.put("DEBUG: Trying to write information collected by buffering service to database")
            sleep(1)
            # self.queue_logging.put(self.__htcondor_dict)
            # self.queue_logging.put(self.__xrootd_dict)
            self.queue_logging.put("DEBUG: Successfully wrote information collected by buffering service to database")
        except Exception as err:
            self.queue_logging.put("ERROR: Failed to write to the database\n" + "Python-ERROR: " + str(err))
        return

    def client_processes_write_queue(self):
        # TODO: Move this part to the NaviX main process to be available to all ClientProcesses
        pass
