from navix.replygenerator.check_time_stamp import check_timestamp
from navix.replygenerator.generate_file_list import create_file_list_to_search
from navix.replygenerator.locate_files import get_location_of_files
from navix.replygenerator.create_reply import create_hpda_rank,\
    add_timestamp,\
    add_hpda_requirement,\
    add_classad_for_monitoring_to_reply
from navix.replygenerator.check_waiting_time import check_waiting_time_and_add_resources
from navix.replygenerator.htcondor_info import get_htcondor_information


#############################################################################################################
#
# Every operation performed on received data, is started from here
# -> except for final hook!
#
#############################################################################################################


class ComputeAndGenerateReply:
    def __init__(self, data, queue_logging, thread_num, settings):
        """
        :param data:
        :param queue_logging:
        :param thread_num:
        initializes every used thread global variable and calls afterwards the clientExec
        """
        self.settings = settings
        self.thread_num = thread_num
        self.queue_logging = queue_logging
        self.data = data
        self.last_time = False
        return

    def compute_generate_reply(self):
        # Process Data
        """
        :param self:
        :return:
        this function coordinates whats happen with the received data and generates a reply which will be sent back via
        socket to the hook
        """
        self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Start processing received data")

        # check input data for timestamp
        try:
            self.data, self.last_time = check_timestamp(self.queue_logging, self.thread_num, self.data)
        except Exception as err:
            raise ValueError('Failed to check timestamp!'
                             + 'ERROR: ' + str(err))

        # fill list with requested files for further processing -> e.g. searching the files via pyxrootd
        try:
            files_to_search_array, input_tmp = create_file_list_to_search(self.queue_logging, self.thread_num, self.data)
        except Exception as err:
            raise ValueError('Failed to cast required files into a list!'
                             + 'ERROR: ' + str(err))

        self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Start requesting file location from xrootd")

        # get location of files
        try:
            count_found_files, server_dict, files_monitor_dict = \
                get_location_of_files(self.queue_logging, self.thread_num, files_to_search_array, self.settings)
        except Exception as err:
            raise ValueError('Error while locating files via pyXrootD!'
                             + 'ERROR: ' + str(err))

        # get HTCondor information for advanced coordination of data and job
        try:
            htcondor_dict = get_htcondor_information(self.queue_logging, self.thread_num, self.settings)
        except Exception as err:
            raise ValueError('Error while getting additional information via HTCondor python bindings!'
                             + 'ERROR: ' + str(err))

        self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Start generating response for hook")

        self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Server found with weight and count of files: "
                               + str(server_dict))

        # check how much time went by since last hook -> if it is above settings.time_before_adding_additional_resources
        # a random server will be added to minimize waiting time!
        try:
            check_waiting_time_and_add_resources(self.queue_logging, self.thread_num, server_dict,
                                                 self.last_time, self.settings)
        except Exception as err:
            raise ValueError('Error while either checking the waiting time or while adding additional worker nodes!'
                             + 'ERROR: ' + str(err))

        # get HPDA_RANK
        self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Next step is adding hpda_rank.")
        try:
            reply, monitor_rank_server = create_hpda_rank(self.queue_logging, self.thread_num,
                                                          server_dict, count_found_files, self.settings)
        except Exception as e:
            self.queue_logging.put("ERROR: " + str(self.thread_num)
                                   + ": Failed while creating HPDA_RANK! -> Abort!"
                                   + "\nPython-ERROR: " + str(e))
            raise ValueError('Failed in creating response!')

        # add HPDA_timestamp for monitoring / update hook
        self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Next step is adding hpda_timestamp.")
        try:
            reply = add_timestamp(self.queue_logging, self.thread_num, reply, self.last_time, self.settings)
        except Exception as e:
            self.queue_logging.put("ERROR: " + str(self.thread_num)
                                   + ": Failed while adding HPDA_timestamp! -> Abort!"
                                   + "\nPython-ERROR: " + str(e))
            raise ValueError('Failed while appending HPDA_timestamp! ' + str(e))

        # create response send to hook
        self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Next step is adding hpda_requirement.")
        try:
            reply += add_hpda_requirement(self.queue_logging, self.thread_num, count_found_files, server_dict,
                                          self.last_time, self.settings)
        except Exception as e:
            self.queue_logging.put("ERROR: " + str(self.thread_num)
                                   + ": Failed while creating HPDA_REQUIREMENT! -> Abort!"
                                   + "\nPython-ERROR: " + str(e))
            raise ValueError('Failed in creating response!')

        try:
            reply += "InputFiles_tmp = \""  + str(input_tmp) + "\"\n"
        except Exception as e:
            self.queue_logging.put("ERROR: " + str(self.thread_num)
                                   + ": Failed while adding InputFiles_tmp! -> Abort!"
                                   + "\nPython-ERROR: " + str(e))
            raise ValueError('Failed in creating response!')
        self.queue_logging.put("DEBUG: " + str(self.thread_num) + ": Next step is adding classads for monitoring.")
        if self.thread_num is not "TEST":
            try:
                reply = add_classad_for_monitoring_to_reply(self.settings, reply, files_monitor_dict,
                                                            files_to_search_array, monitor_rank_server)
            except Exception as e:
                self.queue_logging.put("ERROR: " + str(self.thread_num)
                                       + ": Failed while creating classads for monitoring! -> Abort!"
                                       + "\nPython-ERROR: " + str(e))
                raise ValueError('Failed in creating classads for monitoring!')

        self.queue_logging.put("DEBUG: " + str(self.thread_num)
                               + ": Reply which will be sent in next step is: \n" + str(reply))

        return reply
